 /***************************************************************************
  *                                                                         *
  * Filename:    v4tv6pm.c                                                  *
  * Author:      Tabitha Tipper <t.tipper@lancaster.ac.uk>                     *
  * Created:     07/01/04                                                   *
  * Modified:    25/01/04                                                   *
  * Description: A netfilter kernel module, rewrites outgoing packets to    *
  *              from IPv4 to IPv6 before putting them back on the wire.    *
  *                                                                         *
  * (c) 2004 Tabitha Tipper, licensed under the GNU General Public License     *
  *     See the LICENSE file for details.                                   *
  ***************************************************************************/

/* BADCLASS
 *
 *  include/linux/net/in.h      <-- tells us that Class E is bad.
 */

/* 
 * Oh hey heres an idea, don't reinject MANGLE instead to change it from a v4
 * packet to V6, then just NF_ACCEPT
 *
 * WHY DIDN'T I THINK OF THIS FIRST!
 *
 */

/*
 * see /usr/src/linux/include/linux/skbuff.h for struct skbuff info
 * see /usr/src/linux/include/linux/ip.h for iphdr
 * see /usr/src/linux/include/linux/ipv6.h for ipv6hdr
 *
 * see /usr/src/linux/include/linux/in6.h for struct in6_addr
 */

#define DEBUG_YES

/* needed ones */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>


/* For netfilter */
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv6.h>

/* For playing with skbuffs */
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/ipv6.h>

/* maybe */
#include <linux/if_ether.h>

/* definitely but who knows whats in each */
#include <linux/udp.h>
#include <linux/icmp.h>
#include <net/icmp.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <net/route.h>
#include <linux/netfilter_ipv4/ip_tables.h>


/* for versioning */
/*
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif  
*/
/* doesn't do much now, will later */
#include <linux/wrapper.h>

/* my own stuff */
#include "v4tv6pm.h"
#include "../dnsrh/dnsrh.h"

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

MODULE_AUTHOR("Tabitha Tipper <t.tipper@lancaster.ac.uk>");
MODULE_DESCRIPTION("A netfilter module that rewrites outgoing v4 packets to v6");
MODULE_LICENSE("GPL");

/***
 * This function deals with incoming packets and is named in the v4tv6_ops
 * struct
 ***/
static unsigned int
v4tv6_hook(unsigned int hook,
	struct sk_buff **pskb,
	const struct net_device *indev,
	const struct net_device *outdev,
	int (*okfn)(struct sk_buff *))
{
/*       if ((*pskb)->len == 200)
               return NF_DROP;
*/

	/* FIXME: change to if */
	/* FIXME: where do we get src addr from? */

        /* FIXME: The functions from dnsrh.h should be exported into the
	 * kernel namespace.  See if we can use them
	 */

	switch ((*pskb)->protocol) {
		case __constant_htons(ETH_P_IP): { /* IPv4 */

			ip_pair * addr = NULL;
			int err = 0;
			struct iphdr * temp = (*pskb)->nh.iph;

			/* Get destination addr */
			/* printk(KERN_INFO "v4tv6pm: v4 raw: %d\n", temp->daddr); */

			/* setup hash */
			addr = kmalloc(sizeof(ip_pair), GFP_KERNEL);

			for (err = 0; err < 4; err++) 
				addr->v4[err] = 0;

			for (err = 0; err < 16; err++) 
				addr->v6[err] = 0;

			addr->next = NULL;
		
			/* This shamelessly stolen from kernel.h
			 * Thanks for the code.
			 */
			addr->v4[0] = ((unsigned char *) &temp->daddr)[0];
			addr->v4[1] = ((unsigned char *) &temp->daddr)[1];
			addr->v4[2] = ((unsigned char *) &temp->daddr)[2];
			addr->v4[3] = ((unsigned char *) &temp->daddr)[3];

			/* perform lookup */
			err = dnsrh_checkForV6(addr);

			/* There was a v6, rewrite */
			if (0 == err) { 

				struct ipv6hdr * v6 = NULL;

				v6 = kmalloc(sizeof(struct ipv6hdr), GFP_KERNEL);

				/* Change skbuff protocol type */
				(*pskb)->protocol = 
					__constant_htons(ETH_P_IPV6);

				/* create new ipv6hdr */
				v6->priority = 0;
				v6->version = 6;

				v6->flow_lbl[0] = 0;
				v6->flow_lbl[1] = 0;
				v6->flow_lbl[2] = 0;

				v6->payload_len = (temp->tot_len - temp->ihl);

				v6->nexthdr = 0;

				v6->hop_limit = temp->ttl;

				/* do ip_summed too */

				/* Transfer addresses over */

				/* setup new pointer */
			/*	kfree((*pskb)->nh.iph);
				(*pskb)->nh.ipv6hdr = v6; */

				kfree(v6); /* for now */
			} 
			/* let modified packet continue */
			kfree(addr);

			return NF_ACCEPT;
			break;
		}

		case __constant_htons(ETH_P_IPV6): { /* IPv6 */
			/* printk(KERN_INFO "v4tv6pm: ETH_P_IPV6 packet.\n");
			 */

			/*
			printk(KERN_INFO "v4tv6pm: v6: %d,%d,%d,%d", 
					(*pskb)->nh->daddr.in6_u[0], 
					(*pskb)->nh->daddr.in6_u[1],
					(*pskb)->nh->daddr.in6_u[2],
					(*pskb)->nh->daddr.in6_u[3]);
			*/
			
			return NF_ACCEPT;
			break;
		}

		default:
			break;
	}

	/* if we can't spot it for gads sake don't mess with it! */
	printk(KERN_DEBUG "v4tv6pm: protocol: %d\n", (*pskb)->protocol);
	return NF_ACCEPT;

}

/***
 * This struct is used to register our interest for netfilter work
 ***/
static struct nf_hook_ops v4tv6_ops = 
	{ { NULL, NULL }, v4tv6_hook, PF_INET, NF_IP_LOCAL_OUT, 0 };

/***
 * This is the function that kickstarts the module.
 ***/
static int __init v4tv6_init(void) 
{
	int err = 0;


	/* register with NF_IP_LOCAL_OUT */
	err = nf_register_hook(&v4tv6_ops);

	printk(KERN_INFO "v4tv6: IPv4 to IPv6 (outgoing) filter loaded (%d).\n"
			, err); 
	return err;
}

/***
 * This function stops the module.
 ***/
static void __exit v4tv6_exit(void) 
{ 
	/* unregister with NF_IP_LOCAL_OUT */
	nf_unregister_hook(&v4tv6_ops);
	printk(KERN_INFO "v4tv6: IPv4 to IPv6 (outgoing) filter unloaded.\n");
}

module_init(v4tv6_init)
module_exit(v4tv6_exit)
