#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <time.h>
#include <sys/time.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#define BUFFER_SIZE 1000

int main(int argc, char * argv[]) { 
  int fnum = 0;			  /* number to start on */
  int lnum = 0;			  /* number to end on */
  struct timeval bt;	  	  /* the time before the operation */
  struct timeval at;		  /* the time after the operation */
  struct timezone tz;		  /* time zone information */
  struct hostent * host;	  /* details from gethostbyname */
  char * hostname;		  /* hostname we care about */
  char * result;		  /* ip address it resolves to */
  int test = 0;			  /* 0 for normal sys, 1 for test sys */

  int i = 0;
  int j = 0;
  int err = 0;

  if (argc < 4) { 
    printf("incorrect arguments.\n");
    return 1;
  }

  if (0 == strncmp("test", argv[1], 4)) { 
    printf("Running in test system mode, will use AF_INET (%d)\n", test);
    test = 1;
  } else { 
    printf("Running in normal sys. mode, will use AF_INET6 (%d)\n", test);
  }

  fnum = atoi(argv[2]);
  lnum = atoi(argv[3]);

  tz.tz_minuteswest = 0;
  tz.tz_dsttime = 0;

  host = (struct hostent *) malloc(sizeof(struct hostent));
  hostname = (char *) malloc(BUFFER_SIZE);
  result = (char *) malloc(BUFFER_SIZE);

  for (i = 0; (i + fnum) <= lnum; i++) { /* run N tests */
    memset(hostname, 0, BUFFER_SIZE);
    memset(result, 0, BUFFER_SIZE);
    memset(host, 0, (int) sizeof(struct hostent));

    snprintf(hostname, 
	100, 
	"fake%d.v6.localnet", 
	(int) (i + fnum));

    if (1 == test) { 
      gettimeofday(&bt, &tz);
      host = gethostbyname2(hostname, AF_INET);
      gettimeofday(&at, &tz);
    }
    else { 
      gettimeofday(&bt, &tz);
      host = gethostbyname2(hostname, AF_INET6);
      gettimeofday(&at, &tz);
    }

    /* printf diff, assume gettimeofday to be O(1) */
    if (1 == test) {
      printf("%d:%d.%d:%s:%s\n", (i + fnum), 
	  (int) (at.tv_sec - bt.tv_sec), 
	  (long int) (at.tv_usec - bt.tv_usec),
	  hostname,
	  inet_ntop(AF_INET, host->h_addr_list[0], result, BUFFER_SIZE)
	  );
    } else { 
      printf("%d:%d.%d:%s:%s\n", (i + fnum), 
	  (int) (at.tv_sec - bt.tv_sec), 
	  (long int) (at.tv_usec - bt.tv_usec),
	  hostname,
	  inet_ntop(AF_INET6, host->h_addr_list[0], result, BUFFER_SIZE)
	  );
    }

    /* For the joy of not thrashing our poor server into the floor 
     * sleep(1);
     */

  }

  free(hostname);
  free(result);

  return 0;
}
